from datetime import date

from tests import *
from tests.helpers import MyHelper
from tests import tests_helper
import unittest
import pytest
import os
from parasut.client import Client


class TestExample(unittest.TestCase):
    # def test_helper(self):
    #     self.assertEqual(MyHelper.days_ago(date.today()), 0)

    # def test_other_helper(self):
    #     assert tests_helper.list_has(3, [1, 2, 3])

    # def test_pass(self):
    #     self.assertEqual(True, True)

    # @pytest.mark.xfail
    # def test_should_fail(self):
    #     self.assertEqual(False, True)

    def testClientIdEnvVariable(self):
        try:
            CLIENT_ID = os.environ['PARASUT_CLIENT_ID']
        except KeyError:
            CLIENT_ID = None
            self.assertIsNotNone(CLIENT_ID)

    def testClientSecretEnvVariable(self):
        try:
            CLIENT_SECRET = os.environ['PARASUT_CLIENT_SECRET']
        except KeyError:
            CLIENT_SECRET = None
            self.assertIsNotNone(CLIENT_SECRET)

    def testUsernameEnvVariable(self):
        try:
            USERNAME = os.environ['PARASUT_USERNAME']
        except KeyError:
            USERNAME = None
            self.assertIsNotNone(USERNAME)

    def testPasswordEnvVariable(self):
        try:
            PASSWORD = os.environ['PARASUT_PASSWORD']
        except KeyError:
            PASSWORD = None
            self.assertIsNotNone(PASSWORD)

    def testClient00(self):
        with self.assertRaises(TypeError):
            Client()

    def testClient01(self):
        with self.assertRaises(TypeError):
            CLIENT_ID = os.environ['PARASUT_CLIENT_ID']
            Client(client_id=CLIENT_ID)

    def testClient02(self):
        with self.assertRaises(TypeError):
            CLIENT_ID = os.environ['PARASUT_CLIENT_ID']
            CLIENT_SECRET = os.environ['PARASUT_CLIENT_SECRET']
            Client(client_id=CLIENT_ID,
                   client_secret=CLIENT_SECRET)

    def testClient03(self):
        with self.assertRaises(TypeError):
            CLIENT_ID = os.environ['PARASUT_CLIENT_ID']
            CLIENT_SECRET = os.environ['PARASUT_CLIENT_SECRET']
            USERNAME = os.environ['PARASUT_USERNAME']
            Client(client_id=CLIENT_ID,
                   client_secret=CLIENT_SECRET,
                   username=USERNAME)

    def testClient0(self):
        CLIENT_ID = os.environ['PARASUT_CLIENT_ID']
        CLIENT_SECRET = os.environ['PARASUT_CLIENT_SECRET']
        USERNAME = os.environ['PARASUT_USERNAME']
        PASSWORD = os.environ['PARASUT_PASSWORD']

        Client(client_id=CLIENT_ID,
               client_secret=CLIENT_SECRET,
               username=USERNAME,
               password=PASSWORD,
               sandbox=True)

    def testClient1(self):
        CLIENT_ID = os.environ['PARASUT_CLIENT_ID']
        CLIENT_SECRET = os.environ['PARASUT_CLIENT_SECRET']
        USERNAME = os.environ['PARASUT_USERNAME']
        PASSWORD = os.environ['PARASUT_PASSWORD']

        Client(client_id=CLIENT_ID,
               client_secret=CLIENT_SECRET,
               username=USERNAME,
               password=PASSWORD)

    def testClient2(self):
        CLIENT_ID = os.environ['PARASUT_CLIENT_ID']
        CLIENT_SECRET = os.environ['PARASUT_CLIENT_SECRET']
        USERNAME = os.environ['PARASUT_USERNAME']
        PASSWORD = os.environ['PARASUT_PASSWORD']

        Client(client_id=CLIENT_ID,
               client_secret=CLIENT_SECRET,
               username=USERNAME,
               password=PASSWORD,
               base_url='https://www.example.com',
               token_url='https://www.example.com/oauth/token')

    def testIndexInvoice(self):
        CLIENT_ID = os.environ['PARASUT_CLIENT_ID']
        CLIENT_SECRET = os.environ['PARASUT_CLIENT_SECRET']
        USERNAME = os.environ['PARASUT_USERNAME']
        PASSWORD = os.environ['PARASUT_PASSWORD']

        client_obj = Client(client_id=CLIENT_ID,
                            client_secret=CLIENT_SECRET,
                            username=USERNAME,
                            password=PASSWORD,
                            sandbox=True)

        client_obj.initialize()
        client_obj.functions.indexInvoice()
        print(client_obj.companyId)

    def testCreateInvoice(self):
        CLIENT_ID = os.environ['PARASUT_CLIENT_ID']
        CLIENT_SECRET = os.environ['PARASUT_CLIENT_SECRET']
        USERNAME = os.environ['PARASUT_USERNAME']
        PASSWORD = os.environ['PARASUT_PASSWORD']

        client_obj = Client(client_id=CLIENT_ID,
                            client_secret=CLIENT_SECRET,
                            username=USERNAME,
                            password=PASSWORD,
                            sandbox=True)

        client_obj.initialize()

        # You can get request parameters from
        # https://apidocs.parasut.com/#operation/createContact
        data = {
            "data": {
              "type": "contacts",
              "attributes": {
                "email": "user@example.com",
                "name": "string",
                "short_name": "string",
                "contact_type": "person",
                "tax_office": "string",
                "tax_number": "string",
                "district": "string",
                "city": "string",
                "address": "string",
                "phone": "string",
                "fax": "string",
                "is_abroad": True,
                "archived": True,
                "iban": "string",
                "account_type": "customer"
              },
              "relationships": {
                "category": {
                    "data": {
                        "type": "item_categories"
                    }
                },
                "contact_people": {
                  "data": [
                    {
                      "type": "contact_people",
                      "attributes": {
                        "name": "string",
                        "email": "user@example.com",
                        "phone": "string",
                        "notes": "string"
                      }
                    }
                  ]
                }
              }
            }
          }

        obj = client_obj.functions.createContact(data)
        print(obj.json())



from parasut.version import Version
from tests import *
from tests.helpers import *
import unittest
import pytest


class TestVersion(unittest.TestCase):

    def test_set_version(self):
        ver = Version("1.0.0")
        self.assertEqual(ver.number, "1.0.0")

    def test_version_immutable(self):
        ver = Version("1.0.0")
        with pytest.raises(TypeError) as e:
            ver.number = "1.1.0"



from requests_oauthlib import OAuth2Session
from oauthlib.oauth2 import LegacyApplicationClient, TokenExpiredError
import json
from .functions import Functions
from . import urls
# from functools import wraps


class Client(object):
    """Documentation for Client
    """
    def __init__(self, **kwargs):
        self.client_id = kwargs.get('client_id')
        self.client_secret = kwargs.get('client_secret')
        self.username = kwargs.get('username')
        self.password = kwargs.get('password')

        # Sandbox var if production or test
        self.sandbox = False
        if 'sandbox' in kwargs:
            self.sandbox = kwargs.get('sandbox')
        else:
            self.sandbox = False

        # BaseUrl settings for custom url
        self.baseUrl = None
        if 'base_url' in kwargs:
            self.baseUrl = kwargs.get('base_url')
        else:
            if self.sandbox:
                self.baseUrl = urls.SANDBOX_BASE_URL

        # TokenUrl settings if custom url
        self.tokenUrl = None
        if 'token_url' in kwargs:
            self.tokenUrl = kwargs.get('token_url')
        else:
            if self.tokenUrl:
                self.tokenUrl = urls.SANDBOX_TOKEN_URL

        self.token = None
        self.request = None
        self.companyId = None
        self.functions = None

        if 'client_id' not in kwargs:
            raise TypeError("You must provide client_id")

        if 'client_secret' not in kwargs:
            raise TypeError("You must provide client_secret")

        if 'username' not in kwargs:
            raise TypeError("You must provide username")

        if 'password' not in kwargs:
            raise TypeError("You must provide password")
        super(Client, self).__init__()

    def saveToken(self, new_token):
        self.token = new_token

    def authorize(self):
        oauth = OAuth2Session(client=LegacyApplicationClient(
            client_id=self.client_id))

        self.request = oauth

    def getToken(self):
        if not self.request:
            self.authorize()

        if self.tokenUrl:
            tokenUrl = self.tokenUrl
        else:
            if self.sandbox:
                tokenUrl = urls.SANDBOX_TOKEN_URL
            else:
                tokenUrl = urls.TOKEN_URL
        token = self.request.fetch_token(token_url=tokenUrl,
                                         username=self.username,
                                         password=self.password,
                                         client_id=self.client_id,
                                         client_secret=self.client_secret)
        self.saveToken(token)

    def refreshToken(self, token):
        if self.tokenUrl:
            tokenUrl = self.tokenUrl
        else:
            if self.sandbox:
                tokenUrl = urls.SANDBOX_TOKEN_URL
            else:
                tokenUrl = urls.TOKEN_URL
        update_token = self.request.refresh_token(
            token_url=tokenUrl,
            refresh_token=self.token.get('refresh_token'),
            client_id=self.client_id,
            client_secret=self.client_secret)
        self.token.update(update_token)

    def makeRequest(self, url, method, data=None):
        if not self.request:
            self.authorize()

        if not self.token:
            self.getToken()

        if method == 'GET':
            try:
                r = self.request.get(url=url)
            except TokenExpiredError:
                self.refreshToken()
                self.request.get(url=url)
            return r
        if method == 'POST':
            try:
                return self.request.post(
                    url=url,
                    headers={'Content-Type': 'application/json'},
                    data=json.dumps(data))
            except TokenExpiredError:
                self.refreshToken()
                return self.request.post(
                    url=url,
                    headers={'Content-Type': 'application/json'},
                    data=json.dumps(data))
        if method == 'PUT':
            try:
                return self.request.put(
                    url=url,
                    headers={'Content-Type': 'application/json'},
                    data=json.dumps(data))
            except TokenExpiredError:
                self.refreshToken()
                return self.request.put(
                    url=url,
                    headers={'Content-Type': 'application/json'},
                    data=json.dumps(data))
        if method == 'DELETE':
            try:
                return self.request.delete(
                    url=url,
                    headers={'Content-Type': 'application/json'},
                    data=json.dumps(data))
            except TokenExpiredError:
                self.refreshToken()
                return self.request.delete(
                    url=url,
                    headers={'Content-Type': 'application/json'},
                    data=json.dumps(data))
        if method == 'PATCH':
            try:
                return self.request.patch(
                    url=url,
                    headers={'Content-Type': 'application/json'},
                    data=json.dumps(data))
            except TokenExpiredError:
                self.refreshToken()
                return self.request.patch(
                    url=url,
                    headers={'Content-Type': 'application/json'},
                    data=json.dumps(data))

    def response(self, obj):
        return obj.json()

    def getCompanyId(self):
        if not self.request:
            self.authorize()

        if not self.token:
            self.getToken()

        try:
            if self.sandbox:
                url = urls.SANDBOX_ME_URL
            else:
                url = urls.ME_URL
            r = self.request.get(url=url)
        except TokenExpiredError:
            self.refreshToken()
            if self.sandbox:
                url = urls.SANDBOX_ME_URL
            else:
                url = urls.ME_URL
            r = self.request.get(url=url)
        res = json.loads(json.dumps(r.json()))
        cid = None

        try:
            if res['errors']:
                raise KeyError("Not found company_id in /me")
        except KeyError:
            for v in res['included']:
                if v['type'] == 'companies':
                    cid = v['id']

        if cid:
            self.companyId = cid

    def getFunctionsClass(self):
        self.functions = Functions(company_id=self.companyId,
                                   base_url=self.baseUrl,
                                   make_request=self.makeRequest)

    def initialize(self):
        if not self.request:
            self.authorize()

        if not self.token:
            self.getToken()

        if not self.companyId:
            self.getCompanyId()

        if not self.functions:
            self.getFunctionsClass()

    # def initializeDec(func):
    #     @wraps(func)
    #     def wrapper(self, *args):
    #         if not self.request:
    #             self.authorize()

    #         if not self.token:
    #             self.getToken()

    #         if not self.companyId:
    #             self.getCompanyId()

    #         if not self.functions:
    #             self.getFunctionsClass()
    #         return func(self, *args)
    #     return wrapper

    # """### Sales"""

    # # Sales: Invoices
    # @initializeDec
    # def indexInvoice(self):
    #     return self.functions.indexInvoice()

    # @initializeDec
    # def createInvoice(self, data):
    #     return self.functions.createInvoice(data)

    # @initializeDec
    # def showInvoice(self, pk):
    #     return self.functions.showInvoice(pk)

    # @initializeDec
    # def editInvoice(self, data, pk):
    #     return self.functions.editInvoice(data, pk)

    # @initializeDec
    # def deleteInvoice(self, pk):
    #     return self.functions.deleteInvoice(pk)

    # @initializeDec
    # def payInvoice(self, data, pk):
    #     return self.functions.payInvoice(data, pk)

    # @initializeDec
    # def cancelInvoice(self, pk):
    #     return self.functions.cancelInvoice(pk)

    # @initializeDec
    # def recoverInvoice(self, pk):
    #     return self.functions.recoverInvoice(pk)

    # @initializeDec
    # def arciveInvoice(self, pk):
    #     return self.functions.archiveInvoice(pk)

    # @initializeDec
    # def unarchiveInvoice(self, pk):
    #     return self.functions.unarchiveInvoice(pk)

    # @initializeDec
    # def covertEstimateInvoice(self, data):
    #     return self.functions.convertEstimateInvoice(data)

    # # Sales: Contacts (or Expenses Contacts)
    # @initializeDec
    # def indexContact(self):
    #     return self.functions.IndexContact()

    # @initializeDec
    # def createContact(self, data):
    #     return self.functions.createContact(data)

    # @initializeDec
    # def showContact(self, pk):
    #     """ Show Contact endpoint via GET
    #     Args:
    #         data: JSON Request data
    #         contact_id: id parameter of contact for url
    #     """
    #     return self.functions.showContact(pk)

    # @initializeDec
    # def editContact(self, data, pk):
    #     return self.functions.editContact(data, pk)

    # @initializeDec
    # def deleteContact(self, pk):
    #     return self.functions.deleteContact(pk)

    # @initializeDec
    # def transactionsContact(self, data, pk):
    #     return self.functions.transactionsContact(data, pk)

    # @initializeDec
    # def payContact(self, data, pk):
    #     return self.functions.payContact(data, pk)

    # """### Expenses"""

    # # Expenses: Bills
    # @initializeDec
    # def indexBill(self):
    #     return self.functions.indexBill()

    # @initializeDec
    # def createBasicBill(self, data):
    #     return self.functions.createBasicBill(data)

    # @initializeDec
    # def createDetailedBill(self, data):
    #     return self.functions.createDetailedBill(data)

    # @initializeDec
    # def showBill(self, pk):
    #     return self.functions.showBill(pk)



# URL Constants
BASE_URL = 'https://api.parasut.com/v4/'
TOKEN_URL = 'https://api.parasut.com/oauth/token'
SANDBOX_BASE_URL = 'https://api.heroku-staging.parasut.com/v4/'
SANDBOX_TOKEN_URL = 'https://api.heroku-staging.parasut.com/oauth/token'
COMPANY_ID = ':company_id'

# Sales
SALES_INVOICES_URL = BASE_URL + COMPANY_ID + '/sales_invoices'
SALES_INVOICES_URL = BASE_URL + COMPANY_ID + '/sales_invoices'
CONTACTS_URL = BASE_URL + COMPANY_ID + '/contacts'

# Expenses
PURCHASE_BILLS_URL = BASE_URL + COMPANY_ID + '/purchase_bills'
BANK_FEES_URL = BASE_URL + COMPANY_ID + '/bank_fees'
SALARIES_URL = BASE_URL + COMPANY_ID + '/salaries'
TAXES_URL = BASE_URL + COMPANY_ID + '/taxes'
EMPLOYEES_URL = BASE_URL + COMPANY_ID + '/employees'

# LEGALIZE
E_INVOICE_INBOX_URL = BASE_URL + COMPANY_ID + 'e_invoice_inboxes'
E_ARCHIVES_URL = BASE_URL + COMPANY_ID + '/e_archives'
E_INVOICE_URL = BASE_URL + COMPANY_ID + '/e_invoices'
E_SMMS_URL = BASE_URL + COMPANY_ID + '/e_smms'

# CASH
ACCOUNTS_URL = BASE_URL + COMPANY_ID + '/accounts'
TRANSACTIONS_URL = BASE_URL + COMPANY_ID + '/transactions'

# STOCK/HOARD
PRODUCTS_URL = BASE_URL + COMPANY_ID + '/products'
SHIPMENT_DOCUMENTS_URL = BASE_URL + COMPANY_ID + 'shipment_documents'
STOCK_MOVEMENTS_URL = BASE_URL + COMPANY_ID + '/stock_movements'

# SETTINGS
CATEGORIES_URL = BASE_URL + COMPANY_ID + '/item_categories'
TAGS_URL = BASE_URL + COMPANY_ID + '/tags'

# OTHER
ME_URL = BASE_URL + 'me'
SANDBOX_ME_URL = SANDBOX_BASE_URL + 'me'
TRACKABLE_JOB = BASE_URL + COMPANY_ID + '/trackable_jobs/'



import setuptools
from parasut.version import Version

setuptools.setup(name='parasut-python',
                 version=Version('0.0.5').number,
                 description='Parasut API Python Wrapper',
                 long_description=open('README.md').read().strip(),
                 long_description_content_type="text/markdown",
                 author='Burak Karahan',
                 author_email='burak.karahan@mail.ru',
                 url='https://github.com/marlonjd/parasut-python',
                 packages=setuptools.find_packages(include=['parasut']),
                 zip_safe=False,
                 keywords='parasut python api',
                 install_requires=[
                     'pytest>=6.2.1', 'pytest-cov>=2.10.1', 'requests_oauthlib>=1.3.0'
                 ],
                 classifiers=[
                     "Programming Language :: Python :: 3",
                     "License :: OSI Approved :: MIT License",
                     "Operating System :: OS Independent",
                 ],
                 python_requires='>=3.6')



from . import urls


class Functions(object):
    """Functions for API Endpoints

    """
    def __init__(self, **kwargs):
        if 'base_url' in kwargs:
            self.baseUrl = kwargs.get('base_url')
        else:
            self.baseUrl = None
        if 'company_id' not in kwargs:
            raise TypeError("You must provide company_id")
        else:
            self.company_id = kwargs.get('company_id')
        if 'make_request' not in kwargs:
            raise TypeError("You must provide maka_request function)")
        else:
            self.makeRequest = kwargs.get('make_request')
        super(Functions, self).__init__()

    def replaceUrl(self, url):
        url_obj = url.replace(':company_id', self.company_id)
        if self.baseUrl:
            url_obj = url_obj.replace(urls.BASE_URL, self.baseUrl)
        return url_obj

    def indexGeneral(self, uri):
        url = self.replaceUrl(uri)
        return self.makeRequest(url, 'GET')

    def createGeneral(self, uri, data):
        url = self.replaceUrl(uri)
        return self.makeRequest(url, 'POST', data)

    def showGeneral(self, uri, pk):
        url = self.replaceUrl(uri + '/' + pk)
        return self.makeRequest(url, 'GET')

    def editGeneral(self, uri, data, pk):
        url = self.replaceUrl(uri + '/' + pk)
        return self.makeRequest(url, 'PUT', data)

    def deleteGeneral(self, uri, pk):
        url = self.replaceUrl(uri + '/' + pk)
        return self.makeRequest(url, 'DELETE')

    def cancelGeneral(self, uri, pk):
        url = self.replaceUrl(uri + '/' + pk + '/cancel')
        return self.makeRequest(url, 'DELETE')

    def recoverGeneral(self, uri, pk):
        url = self.replaceUrl(uri + '/' + pk + '/recover')
        return self.makeRequest(url, 'PATCH')

    def archiveGeneral(self, uri, pk):
        url = self.replaceUrl(uri + '/' + pk + '/archive')
        return self.makeRequest(url, 'PATCH')

    def unarchiveGeneral(self, uri, pk):
        url = self.replaceUrl(uri + '/' + pk +
                              '/unarchive')
        return self.makeRequest(url, 'PATCH')

    """### Sales ###"""

    # Sales: Invoices
    def indexInvoice(self):
        """Sales: Invoices Index via GET
        """
        return self.indexGeneral(urls.SALES_INVOICES_URL)

    def createInvoice(self, data):
        """Sales: Invoices Create via POST
        """
        return self.createGeneral(urls.SALES_INVOICES_URL, data)

    def showInvoice(self, pk):
        """Sales: Invoices Show via GET
        """
        return self.showGeneral(urls.SALES_INVOICES_URL, pk)

    def editInvoice(self, data, pk):
        """Sales: Invoices Edit via PUT
        """
        return self.editGeneral(urls.SALES_INVOICES_URL, data, pk)

    def deleteInvoice(self, pk):
        """Sales: Invoices Delete via DELETE
        """
        return self.deleteGeneral(urls.SALES_INVOICES_URL, pk)

    def payInvoice(self, data, pk):
        """Sales: Invoices Pay via POST
        """
        url = self.replaceUrl(urls.SALES_INVOICES_URL + '/' + pk + '/payments')
        return self.makeRequest(url, 'POST', data)

    def cancelInvoice(self, pk):
        """Sales: Invoices Cancel via DELETE
        """
        return self.cancelGeneral(urls.SALES_INVOICES_URL, pk)

    def recoverInvoice(self, pk):
        """Sales: Invoices Recover via PATCH
        """
        return self.recoverGeneral(urls.SALES_INVOICES_URL, pk)

    def archiveInvoice(self, pk):
        """Sales: Invoices Archive via PATCH
        """
        return self.archiveGeneral(urls.SALES_INVOICES_URL, pk)

    def unarchiveInvoice(self, pk):
        """Sales: Invoices Unarchive via PATCH
        """
        return self.unarchiveGeneral(urls.SALES_INVOICES_URL, pk)

    def convertEstimateInvoice(self, data, pk):
        """Sales: Invoices Recover via PATCH
        """
        url = self.replaceUrl(urls.SALES_INVOICES_URL + '/' + pk +
                              '/corvert_to_invoice')
        return self.makeRequest(url, 'PATCH', data)

    # Sales: Contacts (or Expenses Contacts)
    def indexContact(self):
        """Sales: Contacts List via GET
        """
        return self.indexGeneral(urls.CONTACTS_URL)

    def createContact(self, data):
        """Sales: Contacts Create via POST
        """
        return self.createGeneral(urls.CONTACTS_URL, data)

    def showContact(self, pk):
        """Sales: Contacts Show via GET
        """
        return self.showGeneral(urls.CONTACTS_URL, pk)

    def editContact(self, data, pk):
        """Sales: Contacts Edit via POST
        """
        return self.editGeneral(urls.CONTACTS_URL, data, pk)

    def deleteContact(self, pk):
        """Sales: Contacts Delete via DELETE
        """
        return self.deleteGeneral(urls.CONTACTS_URL, pk)

    def transactionsContact(self, data, pk):
        """Sales: Contacts Debit Transactions via POST
        """
        url = self.replaceUrl(urls.CONTACTS_URL + '/' + pk +
                              '/contact_debit_transactions')
        return self.makeRequest(url, 'POST', data)

    def payContact(self, data, pk):
        """Sales: Contacts Debit Transactions via POST
        """
        url = self.replaceUrl(urls.CONTACTS_URL + '/' + pk +
                              '/contact_credit_transactions')
        return self.makeRequest(url, 'POST', data)

    """### Expenses ###"""

    # Expenses: Bills
    def indexBill(self):
        """Expenses: Purchase Bills Index via GET
        """
        return self.indexGeneral(urls.PURCHASE_BILLS_URL)

    def createBasicBill(self, data):
        """Expenses: Purchase Bills Create Basic via POST
        """
        url = self.replaceUrl(urls.PURCHASE_BILLS_URL + '#basic')
        return self.makeRequest(url, 'POST', data)

    def createDetailedBill(self, data):
        """Expenses: Purchase Bills Create Detailed via POST
        """
        url = self.replaceUrl(urls.PURCHASE_BILLS_URL + '#detailed')
        return self.makeRequest(url, 'POST', data)

    def showBill(self, pk):
        """Expenses: Purchase Bills Show via GET
        """
        return self.showGeneral(urls.PURCHASE_BILLS_URL, pk)

    def deleteBill(self, pk):
        """Expenses: Purchase Bills DELETE via DELETE
        """
        return self.deleteGeneral(urls.PURCHASE_BILLS_URL, pk)

    def editBasicBill(self, data, pk):
        """Expenses: Purchase Bills Edit Basic via PUT
        """
        url = self.replaceUrl(urls.PURCHASE_BILLS_URL + '/' + pk + '#basic')
        return self.makeRequest(url, 'PUT', data)

    def editDetailedBill(self, data, pk):
        """Expenses: Purchase Bills Edit Detailed via PUT
        """
        url = self.replaceUrl(urls.PURCHASE_BILLS_URL + '/' + pk + '#detailed')
        return self.makeRequest(url, 'PUT', data)

    def payBill(self, data, pk):
        """Expenses: Purchase Bills Pay via POST
        """
        url = self.replaceUrl(urls.PURCHASE_BILLS_URL + '/' + pk + '/payments')
        return self.makeRequest(url, 'POST', data)

    def cancelBill(self, pk):
        """Expenses: Purchase Bills Cancel via DELETE
        """
        return self.cancelGeneral(urls.PURCHASE_BILLS_URL, pk)

    def recoverBill(self, pk):
        """Expenses: Purchase Bills Recover via PATCH
        """
        return self.recoverGeneral(urls.PURCHASE_BILLS_URL, pk)

    def archiveBill(self, pk):
        """Expenses: Purchase Bills Archive via PATCH
        """
        return self.archiveGeneral(urls.PURCHASE_BILLS_URL, pk)

    def unarchiveBill(self, pk):
        """Expenses: Purchase Bills Unarchive via PATCH
        """
        return self.unarchiveGeneral(urls.PURCHASE_BILLS_URL, pk)

    # Expenses: Bank Fees
    def createBankFee(self, data):
        """Expenses: Bank Fees Create via POST
        """
        return self.createGeneral(urls.BANK_FEES_URL, data)

    def showBankFee(self, pk):
        """Expenses: Bank Fees Show via GET
        """
        return self.showGeneral(urls.BANK_FEES_URL, pk)

    def editBankFee(self, data, pk):
        """Expenses: Bank Fees Edit via PUT
        """
        return self.editGeneral(urls.BANK_FEES_URL, data, pk)

    def deleteBankFee(self, pk):
        """Expenses: Bank Fees Delete via DELETE
        """
        return self.deleteGeneral(urls.BANK_FEES_URL, pk)

    def archiveBankFee(self, pk):
        """Expenses: Bank Fees Archive via PATCH
        """
        return self.archiveGeneral(urls.BANK_FEES_URL, pk)

    def unarchiveBankFee(self, pk):
        """Expenses: Bank Fees Unarchive via PATCH
        """
        return self.unarchiveGeneral(urls.BANK_FEES_URL, pk)

    def payBankFee(self, data, pk):
        """Expenses: Bank Fees Pay via POST
        """
        url = self.replaceUrl(urls.BANK_FEES_URL + '/' + pk + '/payments')
        return self.makeRequest(url, 'POST', data)

    # Expenses: Salaries
    def indexSalary(self):
        """Expenses: Salaries Index via GET
        """
        return self.indexGeneral(urls.SALARIES_URL)

    def createSalary(self, data):
        """Expenses: Salaries Create via POST
        """
        return self.createGeneral(urls.SALARIES_URL, data)

    def showSalary(self, pk):
        """Expenses: Salaries Show via GET
        """
        return self.showGeneral(urls.SALARIES_URL, pk)

    def editSalary(self, data, pk):
        """Expenses: Salaries Edit via PUT
        """
        return self.editGeneral(urls.SALARIES_URL, data, pk)

    def deleteSalary(self, pk):
        """Expenses: Salaries Delete via DELETE
        """
        return self.deleteGeneral(urls.SALARIES_URL, pk)

    def archiveSalary(self, pk):
        """Expenses: Salaries Archive via PATCH
        """
        return self.archiveGeneral(urls.SALARIES_URL, pk)

    def unarchiveSalary(self, pk):
        """Expenses: Salaries unarchive via PATCH
        """
        return self.unarchiveGeneral(urls.SALARIES_URL, pk)

    def paySalary(self, data, pk):
        """Expenses: Salaries Pay via POST
        """
        url = self.replaceUrl(urls.SALARIES_URL + '/' + pk + '/payments')
        return self.makeRequest(url, 'POST', data)

    # Expenses: Taxes
    def indexTax(self):
        """Expenses: Taxes Index via GET
        """
        return self.indexGeneral(urls.TAXES_URL)

    def createTax(self, data):
        """Expenses: Taxes Create via POST
        """
        return self.createGeneral(urls.TAXES_URL, data)

    def showTax(self, pk):
        """Expenses: Taxes Show via GET
        """
        return self.showGeneral(urls.TAXES_URL, pk)

    def editTax(self, data, pk):
        """Expenses: Taxes Edit via PUT
        """
        return self.editGeneral(urls.TAXES_URL, data, pk)

    def deleteTax(self, pk):
        """Expenses: Taxes Delete via DELETE
        """
        return self.deleteGeneral(urls.TAXES_URL, pk)

    def archiveTax(self, pk):
        """Expenses: Taxes Archive via PATCH
        """
        return self.archiveGeneral(urls.TAXES_URL, pk)

    def unarchiveTax(self, pk):
        """Expenses: Taxes unarchive via PATCH
        """
        return self.unarchiveGeneral(urls.TAXES_URL, pk)

    def payTax(self, data, pk):
        """Expenses: Taxes Pay via POST
        """
        url = self.replaceUrl(urls.TAXES_URL + '/' + pk + '/payments')
        return self.makeRequest(url, 'POST', data)

    # Expenses: Employees
    def indexEmployee(self):
        """Expenses: Employees Index via GET
        """
        return self.indexGeneral(urls.EMPLOYEES_URL)

    def createEmployee(self, data):
        """Expenses: TaxesEmployees Create via POST
        """
        return self.createGeneral(urls.EMPLOYEES_URL, data)

    def showEmployee(self, pk):
        """Expenses: Employees Show via GET
        """
        return self.showGeneral(urls.EMPLOYEES_URL, pk)

    def editEmployee(self, data, pk):
        """Expenses: Employees Edit via PUT
        """
        return self.editGeneral(urls.EMPLOYEES_URL, data, pk)

    def deleteEmployee(self, pk):
        """Expenses: Employees Delete via DELETE
        """
        return self.deleteGeneral(urls.EMPLOYEES_URL, pk)

    def archiveEmployee(self, pk):
        """Expenses: Employees Archive via PATCH
        """
        return self.archiveGeneral(urls.EMPLOYEES_URL, pk)

    def unarchiveEmployee(self, pk):
        """Expenses: Employees unarchive via PATCH
        """
        return self.unarchiveGeneral(urls.EMPLOYEES_URL, pk)

    """### Legalize ###"""

    # Legalize: E-Invoice Inboxes
    def indexEInvoiceInbox(self, vkn):
        """Legelize: E-Invoice Inboxes Index via GET
        """
        url = self.replaceUrl(urls.E_INVOICE_INBOX_URL + "?vkn=" + vkn)
        return self.makeRequest(url, 'GET')

    # Legalize: E-Archives
    def createEArchive(self, data):
        """Legalize: E-Arhives Create via POST
        """
        return self.createGeneral(urls.E_INVOICE_INBOX_URL, data)

    def showEArchive(self, pk):
        """Legalize: E-Arhives Show via GET
        """
        return self.showGeneral(urls.E_INVOICE_INBOX_URL, pk)

    def showEArchivePDF(self, data, pk):
        """Legalize: E-Arhives Show PDF via GET
        """
        url = self.replaceUrl(urls.E_ARCHIVES_URL + '/' + pk + '/pdf')
        return self.makeRequest(url, 'GET', data)

    # Legalize: E-Invoices
    def createEInvoice(self, data):
        """Legalize: E-Invoices Create via POST
        """
        return self.createGeneral(urls.E_INVOICE_URL, data)

    def showEInvoice(self, pk):
        """Legalize: E-Invoices Show via GET
        """
        return self.showGeneral(urls.E_INVOICE_URL, pk)

    def showEInvoicePDF(self, data, pk):
        """Legalize: E-Invoices Show PDF via GET
        """
        url = self.replaceUrl(urls.E_INVOICE_URL + '/' + pk + '/pdf')
        return self.makeRequest(url, 'GET', data)

    # Legalize: E-SMM
    def createEsmm(self, data):
        """Legalize: E-SMM Create via POST
        """
        return self.createGeneral(urls.E_SMMS_URL, data)

    def showEsmm(self, pk):
        """Legalize: E-SMM Show via GET
        """
        return self.showGeneral(urls.E_SMMS_URL, pk)

    def showEsmmPDF(self, data, pk):
        """Legalize: E-SMM Show PDF via GET
        """
        url = self.replaceUrl(urls.E_SMMS_URL + '/' + pk + '/pdf')
        return self.makeRequest(url, 'GET', data)

    """### Cash ###"""

    # Cash: Accounts
    def indexAccount(self):
        """Cash: Accounts Index via GET
        """
        return self.indexGeneral(urls.ACCOUNTS_URL)

    def createAccount(self, data):
        """Cash: Accounts Create via POST
        """
        return self.createGeneral(urls.ACCOUNTS_URL, data)

    def showAccount(self, pk):
        """Cash: Accounts Show via GET
        """
        return self.showGeneral(urls.ACCOUNTS_URL, pk)

    def editAccount(self, data, pk):
        """Cash: Accounts Edit via PUT
        """
        return self.editGeneral(urls.ACCOUNTS_URL, data, pk)

    def deleteAccount(self, pk):
        """Cash: Accounts Delete via DELETE
        """
        return self.deleteGeneral(urls.ACCOUNTS_URL, pk)

    def transactionsAccount(self, pk):
        """Cash: Accounts Transactions Get via GET
        """
        url = self.replaceUrl(urls.ACCOUNTS_URL + '/' + pk + '/transactions')
        return self.makeRequest(url, 'GET')

    def debitTransactionAccount(self, data, pk):
        """Cash: Accounts Debit Transaction via POST
        """
        url = self.replaceUrl(urls.ACCOUNTS_URL + '/' + pk +
                              '/debit_transactions')
        return self.makeRequest(url, 'POST', data)

    def creditTransactionAccount(self, data, pk):
        """Cash: Accounts Credit Transaction via POST
        """
        url = self.replaceUrl(urls.ACCOUNTS_URL + '/' + pk +
                              '/credit_transactions')
        return self.makeRequest(url, 'POST', data)

    # Cash: Transactions
    def showTransaction(self, pk):
        """Cash: Accounts Show via GET
        """
        return self.showGeneral(urls.ACCOUNTS_URL, pk)

    def deleteTransaction(self, pk):
        """Cash: Accounts Delete via DELETE
        """
        return self.deleteGeneral(urls.ACCOUNTS_URL, pk)

    """### Stock/Hoard ###"""

    # Stock: Products
    def indexProduct(self):
        """Stock: Products Index via GET
        """
        return self.indexGeneral(urls.PRODUCTS_URL)

    def createProduct(self, data):
        """Stock: Products Create via POST
        """
        return self.createGeneral(urls.PRODUCTS_URL, data)

    def showProduct(self, pk):
        """Stock: Products Show via GET
        """
        return self.showGeneral(urls.PRODUCTS_URL, pk)

    def editProduct(self, data, pk):
        """Stock: Products Edit via PUT
        """
        return self.editGeneral(urls.PRODUCTS_URL, data, pk)

    def deleteProduct(self, pk):
        """Stock: Products Delete via DELETE
        """
        return self.deleteGeneral(urls.PRODUCTS_URL, pk)

    # Stock: Shipment Documents
    def indexShipmentDocument(self):
        """Stock: Shipment Documents Index via GET
        """
        return self.indexGeneral(urls.SHIPMENT_DOCUMENTS_URL)

    def createShipmentDocument(self, data):
        """Stock: Shipment Documents Create via POST
        """
        return self.createGeneral(urls.SHIPMENT_DOCUMENTS_URL, data)

    def showShipmentDocument(self, pk):
        """Stock: Shipment Documents Show via GET
        """
        return self.showGeneral(urls.SHIPMENT_DOCUMENTS_URL, pk)

    def editShipmentDocument(self, data, pk):
        """Stock: Shipment Documents Edit via PUT
        """
        return self.editGeneral(urls.SHIPMENT_DOCUMENTS_URL, data, pk)

    def deleteShipmentDocument(self, pk):
        """Stock: Shipment Documents Delete via DELETE
        """
        return self.deleteGeneral(urls.SHIPMENT_DOCUMENTS_URL, pk)

    # Stocks: Stock Movements
    def indexStockMovement(self):
        """Stock: Stock Movements Index via GET
        """
        return self.indexGeneral(urls.STOCK_MOVEMENTS_URL)

    """### Settings ###"""

    # Settings: Item Categories
    def indexItemCategory(self):
        """Settings: Item Categories Index via GET
        """
        return self.indexGeneral(urls.CATEGORIES_URL)

    def createItemCategory(self, data):
        """Settings: Item Categories Create via POST
        """
        return self.createGeneral(urls.CATEGORIES_URL, data)

    def showItemCategory(self, pk):
        """Settings: Item Categories Show via GET
        """
        return self.showGeneral(urls.CATEGORIES_URL, pk)

    def editItemCategory(self, data, pk):
        """Settings: Item Categories Edit via PUT
        """
        return self.editGeneral(urls.CATEGORIES_URL, data, pk)

    def deleteItemCategory(self, pk):
        """Settings: Item Categories Delete via DELETE
        """
        return self.deleteGeneral(urls.CATEGORIES_URL, pk)

    # Settings: Tags
    def indexTag(self):
        """Settings: Tags Index via GET
        """
        return self.indexGeneral(urls.TAGS_URL)

    def createTag(self, data):
        """Settings: Tags Create via POST
        """
        return self.createGeneral(urls.TAGS_URL, data)

    def showTag(self, pk):
        """Settings: Tags Show via GET
        """
        return self.showGeneral(urls.TAGS_URL, pk)

    def editTag(self, data, pk):
        """Settings: Tags Edit via PUT
        """
        return self.editGeneral(urls.TAGS_URL, data, pk)

    def deleteTag(self, pk):
        """Settings: Tags Delete via DELETE
        """
        return self.deleteGeneral(urls.TAGS_URL, pk)

    """### Other ###"""

    # Other: API Home
    def showAPIHome(self, pk):
        """Other: API Home via GET
        """
        return self.showGeneral(urls.ME_URL)

    # Other: Tackable Job
    def showTackableJob(self, pk):
        """Other: Tackable Job Show via GET
        """
        return self.showGeneral(urls.TRACKABLE_JOB, pk)
